// --------------------------------------------------------------------------
// File part of qMini, a minimal 3D app to extend.
//
// Copyright(C) 2009-2013
// Tamy Boubekeur
//                                                                            
// All rights reserved.                                                       
//                                                                            
// This program is free software; you can redistribute it and/or modify       
// it under the terms of the GNU General Public License as published by       
// the Free Software Foundation; either version 2 of the License, or          
// (at your option) any later version.                                        
//                                                                            
// This program is distributed in the hope that it will be useful,            
// but WITHOUT ANY WARRANTY; without even the implied warranty of             
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              
// GNU General Public License (http://www.gnu.org/licenses/gpl.txt)           
// for more details.                                                          
// --------------------------------------------------------------------------
#define GLEW_STATIC 1
#include <GL/glew.h>

#include "GLApp.h"
#include <QCoreApplication>
#include <QMouseEvent>

#include <string>

using namespace std;

void GLViewer::open (const QString & filename) {
    try {
        mesh.loadOFF (string (qPrintable (filename)));
    } catch (Mesh::Exception e) {
        cerr << e.msg () << endl;
    }
    updateGL ();
}

void GLViewer::initLights () {
    GLfloat light_position1[4] = {42, 374, 161, 0};
    GLfloat light_position2[4] = {473, -351, -259, 0};
    GLfloat light_position3[4] = {-438, 167, -48, 0};

    GLfloat direction1[3] = {-42, -374, -161,};
    GLfloat direction2[3] = {-473, 351, 259};
    GLfloat direction3[3] = {438, -167, 48};

    GLfloat color1[4] = {1.0, 1.0, 1.0, 1};
    GLfloat color2[4] = {0.28, 0.39, 1.0, 1};
    GLfloat color3[4] = {1.0, 0.69, 0.23, 1};

    GLfloat specularColor1[4] = {0.8, 0.8, 0.8, 1};
    GLfloat specularColor2[4] = {0.8, 0.8, 0.8, 1};
    GLfloat specularColor3[4] = {0.8, 0.8, 0.8, 1};

    GLfloat ambient[4] = {0.3f, 0.3f, 0.3f, 0.5f};

    glLightfv (GL_LIGHT0, GL_POSITION, light_position1);
    glLightfv (GL_LIGHT0, GL_SPOT_DIRECTION, direction1);
    glLightfv (GL_LIGHT0, GL_DIFFUSE, color1);
    glLightfv (GL_LIGHT0, GL_SPECULAR, specularColor1);

    glLightfv (GL_LIGHT1, GL_POSITION, light_position2);
    glLightfv (GL_LIGHT1, GL_SPOT_DIRECTION, direction2);
    glLightfv (GL_LIGHT1, GL_DIFFUSE, color2);
    glLightfv (GL_LIGHT1, GL_SPECULAR, specularColor2);

    glLightfv (GL_LIGHT2, GL_POSITION, light_position3);
    glLightfv (GL_LIGHT2, GL_SPOT_DIRECTION, direction3);
    glLightfv (GL_LIGHT2, GL_DIFFUSE, color3);
    glLightfv (GL_LIGHT2, GL_SPECULAR, specularColor3);

    glLightModelfv (GL_LIGHT_MODEL_AMBIENT, ambient);

    glEnable (GL_LIGHTING);
    glEnable (GL_LIGHT0);
    glEnable (GL_LIGHT1);
    glEnable (GL_LIGHT2);
    glEnable (GL_LIGHT3);
}

void GLViewer::init () {
  glewInit();
  if (!GLEW_VERSION_2_0) {
      cerr << "Driver does not support OpenGL v2.0" << endl;
      exit (EXIT_FAILURE);
  }
  setBackgroundColor (Qt::white);
  setAutoBufferSwap (true);
  glCullFace (GL_BACK);
  glEnable (GL_CULL_FACE);
  glDepthFunc (GL_LESS);
  glEnable (GL_DEPTH_TEST);
  initLights ();
  try {
      phongShader = new Shader;
      phongShader->loadFromFile ("shaders/phong.vert", "shaders/phong.frag");
      phongShader->bind ();
  } catch (Shader::Exception e) {
      cerr << e.msg () << endl;
      exit (EXIT_FAILURE);
  }
}


QString GLViewer::helpString() const {
  QString text("<h2>qMini: Minimal 3D Viewer</h2>");
  text += "<hr><br>";
  text += "Author : Tamy Boubekeur (http://www.telecom-paristech.fr/~boubek)<br>";
  text += "<hr><br>";
  text += "Usage: ./qMini <file.off> <br>";
  text += "<hr><br>";  
  text += "<hr><br>";
  text += "Keyboard commands<br>";
  text += "<hr><br>";  
  text += " ?: Print help<br>";
  text += " w: Toggle wireframe Mode<br>";
  text += " q, <esc>: Quit<br>";
  text += "<hr><br>";  
  return text;
}

void GLViewer::draw () {
    mesh.draw ();
}

void GLViewer::keyPressEvent (QKeyEvent * e) {
    if (e->text () == "w") {
        GLint polyMode;
        glGetIntegerv (GL_POLYGON_MODE, &polyMode);
        if (polyMode == GL_LINE) 
            glPolygonMode (GL_FRONT_AND_BACK, GL_FILL);
        else 
            glPolygonMode (GL_FRONT_AND_BACK, GL_LINE);
        updateGL ();
    }  else
        QGLViewer::keyPressEvent (e);
}

// ---------------------------------------------------------
// Entry point of the executable.
// ---------------------------------------------------------

int main (int argc, char **argv) {
    QApplication qMini ( argc, argv );
    GLApp * app = new GLApp ("qmini by Tamy Boubekeur", argc, argv);
    qMini.connect (&qMini, SIGNAL (lastWindowClosed()), &qMini, SLOT (quit()));
    return qMini.exec ();
}
