// ------------------------------------------
// Mesh class directly bindable to the GPU.
// Copyright (C) 2013 Tamy Boubekeur
// All rights reserved.
// ------------------------------------------
#pragma once

#include <vector>
#include <cmath>
#include <string>
#include <iostream>
#include <fstream>
#include <GL/gl.h>
#include "Vec3D.h"

class Mesh {
public:
    typedef struct {Vec3Df p; Vec3Df n;} Vertex;
    typedef Vec3D<unsigned int> Triangle;
    class Exception {
    public:
        inline Exception (const std::string & msg) : _msg ("Mesh Exception: " + msg) {}
        inline const std::string & msg () const { return _msg; }
    protected:
        std::string _msg;
    };
    inline std::vector<Vertex> & V () { return _V; }
    inline std::vector<Triangle> & T () { return _T; }
    inline const std::vector<Vertex> & V () const { return _V; }
    inline const std::vector<Triangle> & T () const { return _T; }
    inline void clear () { _V.clear (); _T.clear (); }
    void recomputeNormals () {
        for (unsigned int i = 0; i < _T.size (); i++) {
                Vec3Df e01 (_V[_T[i][1]].p - _V[_T[i][0]].p);
                Vec3Df e02 (_V[_T[i][2]].p - _V[_T[i][0]].p);
                Vec3Df nt = Vec3Df::crossProduct (e01, e02);
                nt.normalize ();
                for (unsigned int j = 0; j < 3; j++) {
                    _V[_T[i][j]].n += nt;
                }
            }
        for (unsigned int i = 0; i < _V.size (); i++)
            _V[i].n.normalize ();
    }
    void draw () {
        glEnableClientState (GL_VERTEX_ARRAY);
        glEnableClientState (GL_NORMAL_ARRAY);
        glVertexPointer (3, GL_FLOAT, sizeof (Vertex), (GLvoid*)(&_V[0].p));
        glNormalPointer (GL_FLOAT, sizeof (Vertex), (GLvoid*)(((float*)&_V[0].p) + 3));
        glDrawElements (GL_TRIANGLES, 3*_T.size(), GL_UNSIGNED_INT, (GLvoid*)(&_T[0]));
        glDisableClientState (GL_VERTEX_ARRAY);
        glDisableClientState (GL_NORMAL_ARRAY);
    }
    void loadOFF (const std::string & filename) {
        clear ();
        std::ifstream in (filename.c_str ());
        if (!in)
            throw Exception ("Error loading OFF file: " + filename);
        std::string offString;
        unsigned int numV, numF, tmp;
        in >> offString >> numV >> numF >> tmp;
        _V.resize (numV);
        for (unsigned int i = 0; i < numV; i++)
            in >> _V[i].p;
        unsigned int s;
        for (unsigned int i = 0; i < numF; i++) {
            in >> s;
            std::vector<unsigned int> v(s);
            for (unsigned int j = 0; j < s; j++)
                in >> v[j];
            for (unsigned int j = 2; j < s; j++)
                _T.push_back (Triangle (v[j-2], v[j-1], v[j]));
        }
        in.close ();
        recomputeNormals();
    }
private:
    std::vector<Vertex> _V;
    std::vector<Triangle> _T;
};

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
